﻿<# 
.SYNOPSIS 
.DESCRIPTION 
.PARAMETER out
.PARAMETER storageName
.PARAMETER storageAccessKey
.PARAMETER projectName
.PARAMETER projectVersion
.PARAMETER vmsize
.PARAMETER workerRoleName
.PARAMETER zulu
.NOTES 
    Author: Ksisu, 2016. 10. 26. http://wanari.com
#> 
Param(
  [Parameter(Mandatory=$true)][string]$out = "out.zip",
  [Parameter(Mandatory=$true)][string]$storageName,
  [Parameter(Mandatory=$true)][string]$storageAccessKey,
  [Parameter(Mandatory=$true)][string]$projectName,
  [Parameter(Mandatory=$true)][string]$projectVersion,
  [Parameter(Mandatory=$true)][string]$vmSize="Small",
  [Parameter(Mandatory=$true)][string]$workerRoleName="WorkerRole1",
  [Parameter(Mandatory=$true)][string]$zulu="zulu8.17.0.3-jdk8.0.102-win_x64"
)

# https://azure.microsoft.com/hu-hu/documentation/articles/cloud-services-sizes-specs/
$VM_avaible = @( 
    "A0", "ExtraSmall",
    "A1", "Small",
    "A2", "Medium",
    "A3", "Large",
    "A4", "ExtraLarge",
    "A5",
    "A7",
    "A8",
    "A9",
    "A10",
    "A11",
    "Standard_D1",
    "Standard_D2",
    "Standard_D3",
    "Standard_D4",
    "Standard_D11",
    "Standard_D12",
    "Standard_D13",
    "Standard_D14",
    "Standard_D1_v2",
    "Standard_D2_v2",
    "Standard_D3_v2",
    "Standard_D4_v2",
    "Standard_D5_v2",
    "Standard_D11_v2",
    "Standard_D12_v2",
    "Standard_D13_v2",
    "Standard_D14_v2",
    "Standard_D15_v2"
    )

# Check VM size
if ($VM_avaible -notcontains $vmsize){
  throw [System.ArgumentException] "Not valid VM size. Check: https://azure.microsoft.com/hu-hu/documentation/articles/cloud-services-sizes-specs/"
}

# Check Zulu JDK
$zulu_check = $null
Try {
  $zulu_check = Invoke-WebRequest "http://azure.azulsystems.com/zulu/$zulu.zip" -DisableKeepAlive -UseBasicParsing -Method head
} Catch {
}
if(($zulu_check -eq $null) -or ($zulu_check.StatusCode -ne 200)){
  throw [System.ArgumentException] "Not found Zulu JDK. Check http://www.azul.com/downloads/zulu/zulu-windows/"
}

# Delete destination folder if exists
if ((Test-Path .\out) -eq $True){
  Remove-Item -Recurse -Force .\out
}

# Copy template files
Copy-Item .\template .\out -recurse

# Rename WorkerRole folder
Rename-Item .\out\WorkerRole1 $workerRoleName

# Temp variables
$appZip = "$projectName-$projectVersion.zip"
$appRunFolder = "$projectName-$projectVersion\bin"
$appRunScript = "$projectName.bat"

# Set template variables
(Get-Content .\out\$workerRoleName\.startup.cmd) `
  -replace '<<<WORKER_ROLE_NAME>>>', $workerRoleName `
  -replace '<<<STORAGE_NAME>>>', $storageName `
  -replace '<<<STORAGE_ACCESS_KEY>>>', $storageAccessKey `
  -replace '<<<PLAY_APP_ZIP_NAME>>>', $appZip `
  -replace '<<<PLAY_APP_FOLDER_NAME>>>', $appRunFolder `
  -replace '<<<PLAY_APP_START_BAT_NAME>>>', $appRunScript `
  -replace '<<<ZULU_JDK>>>', $zulu `
  | Set-Content .\out\$workerRoleName\.startup.cmd

(Get-Content .\out\ServiceDefinition.csdef) `
  -replace '<<<WORKER_ROLE_NAME>>>', $workerRoleName `
  -replace '<<<VM_SIZE>>>', $vmSize `
  | Set-Content .\out\ServiceDefinition.csdef
  
# Create zip
Compress-Archive -Path .\out\* -DestinationPath .\$out

# Cleaning
Remove-Item -Recurse -Force .\out